/*
 * These routines manage AXI DMA.
 */
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/slab.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/pci.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/miscdevice.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/cdev.h>

#include <linux/ioctl.h>
#include <linux/types.h>

#include "hdlc-drv.h"
#include "axidma.h"
#include "dbg.h"

#if 1

/* Descriptors defines for Tx and Rx DMA - 2^n for the best performance */

#ifdef _DEBUG_
void print_bd( struct axidma_bd *bdp )
{
	printk("**************************************************\n");
	printk("next :  0x%X \tNext BD addr.\n",	bdp->next );
//      printk("reserved1 : 0x%X \n",			bdp->reserved1 );
	printk("phys :  0x%X \tAddr of Data ptr.\n",	bdp->phys );
//      printk("reserved2 : 0x%x \n",   bdp->reserved2 );
//      printk("reserved3 : 0x%x \n",   bdp->reserved3);
//      printk("reserved4 : 0x%x \n",   bdp->reserved4);
	printk("cntrl : 0x%08X \t0x08~:SOF, 0x04~:EOF, Under:Buf_Size.\n",	bdp->cntrl);
	printk("status: 0x%08X \t0x80~:Cmplt, 0x70:ERR_Chk, SOF,EOF, Under:TX_size.\n", bdp->status);
//      printk("app0 :  0x%x \n",	bdp->app0);
//      printk("app1 :  0x%x \n",	bdp->app1);
//      printk("app2 :  0x%x \n",	bdp->app2);
	printk("app3 :  0x%X \tPort ID.\n",	bdp->app3);
	printk("app4 :  0x%08X \tTX Length.\n",	bdp->app4);
	printk("sw_id_offset: 0x%llX\n",	bdp->sw_id_offset);
//      printk("reserved5 : 0x%x \n",   bdp->reserved5);
//      printk("reserved6 : 0x%x \n",   bdp->reserved6);
	printk("**************************************************\n");
	printk("\n");
}
#endif


void axienet_dma_bd_release(struct pcidrv_port *portdev)
{
	int i;

	struct pcidrv_port  *lp = portdev;
	struct pcidrv *pdrv = lp->pcidrv;

	printk("__%s++\n", __FUNCTION__);

	for (i = 0; i < TX_BD_NUM; i++) {
		dma_free_coherent( &pdrv->pcidev->dev, lp->max_frm_size,
				(void *)lp->tx_bd_v[i].sw_id_offset,
				(dma_addr_t) lp->tx_bd_v[i].phys );
	}

	for (i = 0; i < RX_BD_NUM; i++) {
		dma_free_coherent( &pdrv->pcidev->dev, lp->max_frm_size,
				(void *) lp->rx_bd_v[i].sw_id_offset,
				(dma_addr_t) lp->rx_bd_v[i].phys );
	}

	if (lp->rx_bd_v)
		dma_free_coherent( &pdrv->pcidev->dev, sizeof(*lp->rx_bd_v) * RX_BD_NUM,
				lp->rx_bd_v, lp->rx_bd_p );
	if (lp->tx_bd_v) 
		dma_free_coherent( &pdrv->pcidev->dev, sizeof(*lp->tx_bd_v) * TX_BD_NUM,
				lp->tx_bd_v, lp->tx_bd_p );

	printk("__%s++\n", __FUNCTION__);
}


int axienet_dma_bd_init(struct pcidrv_port * portdev)
{
	u32 cr;
	int i,n;

	struct pcidrv_port  *lp = portdev;
	struct pcidrv *pdrv = lp->pcidrv;
	//char *data = NULL;
	port_data_t *data = NULL;

	printk("__%s++\n", __FUNCTION__);
	n = lp->port_id;

	/* Reset the indexes which are used for accessing the BDs */
	lp->tx_bd_ci = 0;
	lp->tx_bd_tail = 0;
	lp->rx_bd_ci = 0;

	/* Allocate the Tx and Rx buffer descriptors. */
	lp->tx_bd_v = dma_alloc_coherent( &pdrv->pcidev->dev,
			sizeof(*lp->tx_bd_v) * TX_BD_NUM, &lp->tx_bd_p, GFP_KERNEL);
	if( !lp->tx_bd_v ) goto out;

	lp->rx_bd_v = dma_alloc_coherent(&pdrv->pcidev->dev,
			sizeof(*lp->rx_bd_v) * RX_BD_NUM, &lp->rx_bd_p, GFP_KERNEL);
	if( !lp->rx_bd_v ) goto out;

#ifdef _DEBUG_
        printk("DBG> lp->tx_bd_p Phys Addr: 0x%llX \n", lp->tx_bd_p );
        printk("DBG> lp->rx_bd_p Phys Addr: 0x%llX \n", lp->rx_bd_p );
#endif
	for( i=0; i < TX_BD_NUM; i++) {
		lp->tx_bd_v[i].next = lp->tx_bd_p + sizeof(*lp->tx_bd_v) * ((i+1) % TX_BD_NUM);

		/* Ensure that the skb is completely updated prio to mapping the DMA */
		wmb();
		data = dma_alloc_coherent( &pdrv->pcidev->dev, lp->max_frm_size ,
				 (dma_addr_t *) &lp->tx_bd_v[i].phys, GFP_ATOMIC |__GFP_ZERO );
		if(!data) goto out;

		lp->tx_bd_v[i].sw_id_offset = (phys_addr_t) data;
		lp->tx_bd_v[i].cntrl = lp->max_frm_size;
	}

	for( i=0; i < RX_BD_NUM; i++) {
		lp->rx_bd_v[i].next = lp->rx_bd_p + sizeof(*lp->rx_bd_v) * ((i+1) % RX_BD_NUM);

		/* Ensure that the skb is completely updated prio to mapping the DMA */
		wmb();
		data = dma_alloc_coherent( &pdrv->pcidev->dev, lp->max_frm_size ,
				 (dma_addr_t *) &lp->rx_bd_v[i].phys, GFP_ATOMIC |__GFP_ZERO );
		if(!data) goto out;

		lp->rx_bd_v[i].sw_id_offset = (phys_addr_t) data;
		//printk("===> 0x%x \n",lp->rx_bd_v[i].sw_id_offset);

		lp->rx_bd_v[i].cntrl = lp->max_frm_size;
	}

	/* Start updating the Rx channel control register */
	cr = axienet_dma_in32( lp, n, XAXIDMA_RX_CR_OFFSET );

	/* Update the interrupt coalesce count */
	cr = ((cr & ~XAXIDMA_COALESCE_MASK) | ((lp->coalesce_count_rx) << XAXIDMA_COALESCE_SHIFT));

	/* Update the delay timer count */
	cr = ((cr & ~XAXIDMA_DELAY_MASK) | (XAXIDMA_DFT_RX_WAITBOUND << XAXIDMA_DELAY_SHIFT));

	/* Enable coalesce, delay timer and error interrupts */
	cr |= XAXIDMA_IRQ_ALL_MASK;

#ifdef _DEBUG_
                printk("DBG> Update DMA_RX_CR: 0x%08X \n", cr );
#endif

	/* Write to the Rx channel control register */
	axienet_dma_out32( lp, n, XAXIDMA_RX_CR_OFFSET, cr );

	/* Start updating the Tx channel control register */
	cr = axienet_dma_in32( lp, n, XAXIDMA_TX_CR_OFFSET );

	/* Update the interrupt coalesce count */
	cr = (((cr & ~XAXIDMA_COALESCE_MASK)) | ((lp->coalesce_count_tx) << XAXIDMA_COALESCE_SHIFT));

	/* Update the delay timer count */
	cr = (((cr & ~XAXIDMA_DELAY_MASK)) | (XAXIDMA_DFT_TX_WAITBOUND << XAXIDMA_DELAY_SHIFT));

	/* Enable coalesce, delay timer and error interrupts */
	cr |= XAXIDMA_IRQ_ALL_MASK;

#ifdef _DEBUG_
                printk("DBG> Update DMA_TX_CR: 0x%08X.\n", cr );
#endif

	/* Write to the Tx channel control register */
	axienet_dma_out32( lp, n, XAXIDMA_TX_CR_OFFSET, cr);

	/* Populate the tail pointer and bring the Rx Axi DMA engine out of halted state.
	 * This will make the Rx side ready for reception. */
	axienet_dma_bdout( lp, n, XAXIDMA_RX_CDESC_OFFSET, lp->rx_bd_p);
	cr = axienet_dma_in32( lp, n, XAXIDMA_RX_CR_OFFSET);
	axienet_dma_out32( lp, n, XAXIDMA_RX_CR_OFFSET, cr | XAXIDMA_CR_RUNSTOP_MASK);
	axienet_dma_bdout( lp, n, XAXIDMA_RX_TDESC_OFFSET, 
			lp->rx_bd_p + ( sizeof(*lp->rx_bd_v) * (RX_BD_NUM-1) ) );

	/* Write to the RS (Run-stop) bit in the Tx channel control register.
	 * Tx channel is now ready to run. But only after we write to the tail pointer register
	 * that the Tx channel will start transmitting. */
	axienet_dma_bdout( lp, n, XAXIDMA_TX_CDESC_OFFSET, lp->tx_bd_p);
	cr = axienet_dma_in32( lp, n, XAXIDMA_TX_CR_OFFSET);
	axienet_dma_out32( lp, n, XAXIDMA_TX_CR_OFFSET, cr | XAXIDMA_CR_RUNSTOP_MASK);

	printk("__%s--\n", __FUNCTION__);
	return 0;

out:
	axienet_dma_bd_release(portdev);
	return -ENOMEM;
}



static inline int axienet_check_tx_bd_space( struct pcidrv_port *lp)
{
	struct axidma_bd *cur_p;

	cur_p = &lp->tx_bd_v[(lp->tx_bd_tail) % TX_BD_NUM];

	//printk("===> 0x%x %x cur_p->status %d\n", 
	//	cur_p->status, cur_p->status & XAXIDMA_BD_STS_ALL_MASK, lp->tx_bd_tail);

	if( cur_p->status & XAXIDMA_BD_STS_ALL_MASK ) return -1;
	if( cur_p->cntrl & XAXIDMA_BD_CTRL_ALL_MASK ) return -1;

	return 0;
}


/* It starts DMA transmition. return 0: OK. */
int axienet_start_xmit( struct pcidrv_port *lp, const char *user, uint32_t len )
{
	dma_addr_t tail_p;
	struct axidma_bd *cur_p;
	port_data_t *data;
	int prtid = lp->port_id;

	if( lp->max_frm_size < len) {
                printk("[%s]WARN: TX_Max_Frame_Size(%d) < User_Frame_Size(%d).\n",
                        __func__, lp->max_frm_size, len );
//              return -1;
        }

	spin_lock_irq( &lp->tx_lock);
	rmb();
	cur_p = &lp->tx_bd_v[lp->tx_bd_tail];

	if( (cur_p->status & XAXIDMA_BD_STS_ALL_MASK) ||
            (cur_p->cntrl & XAXIDMA_BD_CTRL_ALL_MASK) ) {
                spin_unlock_irq(&lp->tx_lock);
                return -1;
        }

	data = (port_data_t *) cur_p->sw_id_offset;
	copy_from_user( (void*)(data->data), user, len);

	cur_p->cntrl = len | XAXIDMA_BD_CTRL_TXSOF_MASK;
	cur_p->cntrl |= XAXIDMA_BD_CTRL_TXEOF_MASK;

	cur_p->app4 = len;
	cur_p->app3 = prtid;

	tail_p = lp->tx_bd_p + sizeof(*lp->tx_bd_v) * lp->tx_bd_tail;

	/* Ensure BD write before starting transfer */
	wmb();

	/* Start the transfer */
	axienet_dma_bdout( lp, prtid, XAXIDMA_TX_TDESC_OFFSET, tail_p);

#ifdef _DEBUG_
        printk("[%s]DBG> Tail_idx:%d, tail_p:0x%llX.\n", __func__, lp->tx_bd_tail, tail_p );
        printk("[%s]DBG> TX_user_len: %d, BD frame size: %d.\n", __func__, len, lp->max_frm_size );
        print_bd( cur_p );
//      print_buff( data->data, len );
#endif

	++lp->tx_bd_tail;
	lp->tx_bd_tail %= TX_BD_NUM;

	spin_unlock_irq( &lp->tx_lock);
	return 0;
}



static int axienet_start_xmit_done( struct pcidrv_port *lp )
{
	u32 packets = 0;
	struct axidma_bd *cur_p;
	unsigned int status = 0;

	rmb();
	cur_p = &lp->tx_bd_v[lp->tx_bd_ci];

#ifdef _DEBUG_
        /* cur_p->status saves actual transferred bytes using XAXIDMA_BD_STS_ACTUAL_LEN_MASK, 
         * 0x07FF_FFFF, and app4 stores user length.
         * XAXIDMA_BD_STS_COMPLETE_MASK = 0x8000_0000. */
        printk( "[%s]DBG> cur_p <-BD[%d]..\n", __func__, lp->tx_bd_ci );
        print_bd( cur_p );
#endif

	status = cur_p->status;
	while( status & XAXIDMA_BD_STS_COMPLETE_MASK) {

		wmb();
//		cur_p->phys = 0;
		cur_p->cntrl = 0;
		cur_p->status = 0;
		cur_p->app0 = 0;
		cur_p->app1 = 0;
		cur_p->app2 = 0;
		cur_p->app4 = 0;

		packets++;
		++lp->tx_bd_ci;
		lp->tx_bd_ci %= TX_BD_NUM;

		rmb();
		cur_p = &lp->tx_bd_v[lp->tx_bd_ci];
		status = cur_p->status;
	}

	return packets;
}


/* TX task Handler */
void xaxienet_tx_poll( unsigned long data)
{
	u32 cr, status;
	struct pcidrv_port  *lp = (struct pcidrv_port *) data;
	int n = lp->port_id;
	int work_done = 0, quota = TX_BD_NUM;

	spin_lock(&lp->tx_lock);
	status = axienet_dma_in32(lp, n, XAXIDMA_TX_SR_OFFSET);

	while( (status & (XAXIDMA_IRQ_IOC_MASK | XAXIDMA_IRQ_DELAY_MASK)) && (work_done<quota) ) {
		axienet_dma_out32( lp, n, XAXIDMA_TX_SR_OFFSET, status);
		work_done += axienet_start_xmit_done(lp);
		status = axienet_dma_in32( lp, n, XAXIDMA_TX_SR_OFFSET);
	}
	spin_unlock(&lp->tx_lock);

#ifdef _DEBUG_
        printk("[%s]DBG> (Work done: %d < quota: %d) is right..\n", __func__, work_done, quota );
        if (work_done == quota) {
                printk("Warning: Quota is over.. tx_poll reschedule");
//              tasklet_schedule(&lp->tx_poll);
        }
#endif

	/* Enable the interrupts again */
	cr = axienet_dma_in32(lp, n, XAXIDMA_TX_CR_OFFSET);

#ifdef _DEBUG_
        printk("[%s]DBG> Interpt again TX_CR Value:0x%08X.\n", __func__, cr );
#endif

	cr |= (XAXIDMA_IRQ_IOC_MASK | XAXIDMA_IRQ_DELAY_MASK);
	axienet_dma_out32(lp, n, XAXIDMA_TX_CR_OFFSET, cr);
}



/* It's a TX interpt handler */
int axienet_tx_irq( void *portdev )
{
	u32 cr;
	unsigned int status;

	struct pcidrv_port *lp = (struct pcidrv_port *) portdev;
	int n = lp->port_id;
	struct pcidrv *pdrv = lp->pcidrv;

	status = axienet_dma_in32(lp, n, XAXIDMA_TX_SR_OFFSET);
#ifdef _DEBUG_
        printk("[%s]DBG> TX_DMASR: 0x%08X.\n", __func__, status );
#endif

// __TASKLET__
#if 0 
	if( status & (XAXIDMA_IRQ_IOC_MASK | XAXIDMA_IRQ_DELAY_MASK) ) {
		cr = axienet_dma_in32( lp, n, XAXIDMA_TX_CR_OFFSET);
		cr &= ~(XAXIDMA_IRQ_IOC_MASK | XAXIDMA_IRQ_DELAY_MASK);

		axienet_dma_out32( lp, n, XAXIDMA_TX_CR_OFFSET, cr);
		tasklet_schedule( &lp->tx_poll);
	}
#else 
	if (status & (XAXIDMA_IRQ_IOC_MASK | XAXIDMA_IRQ_DELAY_MASK)) {
		axienet_dma_out32(lp, n, XAXIDMA_TX_SR_OFFSET, status);
		axienet_start_xmit_done(lp);
		goto out;	
	}
#endif
	if( !(status & XAXIDMA_IRQ_ALL_MASK) )
		dev_err( &pdrv->pcidev->dev, "No interrupts asserted in Tx path");

	if (status & XAXIDMA_IRQ_ERROR_MASK) {
		dev_err( &pdrv->pcidev->dev, "DMA Tx error 0x%x\n", status);
		dev_err( &pdrv->pcidev->dev, "Current BD is at: 0x%x\n",
			(lp->tx_bd_v[lp->tx_bd_ci]).phys );

		cr = axienet_dma_in32( lp, n, XAXIDMA_TX_CR_OFFSET);
		/* Disable coalesce, delay timer and error interrupts */
		cr &= (~XAXIDMA_IRQ_ALL_MASK);
		/* Write to the Tx channel control register */
		axienet_dma_out32( lp, n, XAXIDMA_TX_CR_OFFSET, cr);

		cr = axienet_dma_in32( lp, n, XAXIDMA_RX_CR_OFFSET);
		/* Disable coalesce, delay timer and error interrupts */
		cr &= (~XAXIDMA_IRQ_ALL_MASK);
		/* Write to the Rx channel control register */
		axienet_dma_out32( lp, n, XAXIDMA_RX_CR_OFFSET, cr);

		tasklet_schedule( &lp->dma_err_tasklet );
		axienet_dma_out32( lp, n, XAXIDMA_TX_SR_OFFSET, status);
	}
out:
	return 0;
}


/* It's a RX interpt handler */
int axienet_rx_irq(void *portdev) 
{
	u32 cr;
	unsigned int status;
	struct pcidrv_port *lp = (struct pcidrv_port *) portdev;

	struct pcidrv *pdrv = lp->pcidrv;
	int n = lp->port_id;

	status = axienet_dma_in32( lp, n, XAXIDMA_RX_SR_OFFSET);
	if( status & (XAXIDMA_IRQ_IOC_MASK | XAXIDMA_IRQ_DELAY_MASK) ) {
		cr = axienet_dma_in32( lp, n, XAXIDMA_RX_CR_OFFSET );

#ifdef _DEBUG_
        printk("[%s]DBG> RX_DMASR:0x%08X, DMACR:0x%08X.\n", __func__, status, cr );
        printk("[%s]DBG>(IOC_IRQ=0x1000, DELAY_IRQ=0x2000)\n", __func__ );
#endif
		cr &= ~(XAXIDMA_IRQ_IOC_MASK | XAXIDMA_IRQ_DELAY_MASK);

		axienet_dma_out32( lp, n, XAXIDMA_RX_CR_OFFSET, cr );
		tasklet_schedule( &lp->rx_poll );
	}

	if( !(status & XAXIDMA_IRQ_ALL_MASK) )
		dev_err( &pdrv->pcidev->dev, "No interrupts asserted in Rx path");

	if( status & XAXIDMA_IRQ_ERROR_MASK ) {
		dev_err( &pdrv->pcidev->dev, "DMA Rx error 0x%x\n", status);
		dev_err( &pdrv->pcidev->dev, "Current BD is at: 0x%x\n", 
			(lp->rx_bd_v[lp->rx_bd_ci]).phys);

		cr = axienet_dma_in32( lp, n, XAXIDMA_TX_CR_OFFSET);

		/* Disable coalesce, delay timer and error interrupts */
		cr &= (~XAXIDMA_IRQ_ALL_MASK);
		/* Finally write to the Tx channel control register */
		axienet_dma_out32( lp, n, XAXIDMA_TX_CR_OFFSET, cr);

		cr = axienet_dma_in32( lp,n, XAXIDMA_RX_CR_OFFSET);
		/* Disable coalesce, delay timer and error interrupts */
		cr &= (~XAXIDMA_IRQ_ALL_MASK);
		/* write to the Rx channel control register */
		axienet_dma_out32( lp, n, XAXIDMA_RX_CR_OFFSET, cr);

		tasklet_schedule( &lp->dma_err_tasklet);
		axienet_dma_out32( lp, n, XAXIDMA_RX_SR_OFFSET, status);
	}

	return 0;
}


/* It copies several data frames of BDs into lp->rx_data[widx] location.
 * budget : the number of remained (free) BD */
int axienet_recv( struct pcidrv_port *lp, int budget )
{
	u32 length;
	u32 packets = 0;
	dma_addr_t tail_p = 0;
	struct axidma_bd *cur_p;
	unsigned int numbdfree = 0;

	port_data_t *rxdata = NULL;
	int n = lp->port_id;

	/* Get relevat BD status value */
	rmb();
	cur_p = &lp->rx_bd_v[lp->rx_bd_ci];
	if( !cur_p ) return 0;

	/* At the rx_bd_ci's BD, if its status includes COMPLETE_MASK=0x80~,
         * this function reads the DB's Frame Data, and continue..  */
	while( (numbdfree < budget) && (cur_p->status & XAXIDMA_BD_STS_COMPLETE_MASK) )
       	{
		tail_p = lp->rx_bd_p + sizeof(*lp->rx_bd_v) * lp->rx_bd_ci;
		rmb();
		rxdata = (port_data_t *)(cur_p->sw_id_offset);

		length = cur_p->status & XAXIDMA_BD_STS_ACTUAL_LEN_MASK;

#ifdef _DEBUG_
        printk("[%s]DBG> cur_p <-BD[%d]..\n", __func__, lp->rx_bd_ci );
        print_bd( cur_p );
        print_buff( rxdata->data, length );
#endif

		wmb();
		pci_rx_pkt_receive( lp, rxdata, length);
		packets++;

		/* Ensure that the skb is completely updated prio to mapping the DMA */
		rmb();
		cur_p->status = 0;
		++lp->rx_bd_ci;
		lp->rx_bd_ci %= RX_BD_NUM;
		
		rmb();
		cur_p = &lp->rx_bd_v[lp->rx_bd_ci];
		numbdfree++;
	}

	/* After it reads data from BD, it lets the AXI engine recv the next data */
	if(tail_p) axienet_dma_bdout( lp, n, XAXIDMA_RX_TDESC_OFFSET, tail_p);

	return numbdfree;
}


/* when some data are received, this handler is called and scheduled by the axienet_rx_irq(). */
void xaxienet_rx_poll( unsigned long data)
{
	u32 cr, status;
	int work_done = 0, quota = RX_BD_NUM;
	struct pcidrv_port *lp = (struct pcidrv_port *) data;

	struct pcidrv *pdrv = lp->pcidrv;
	int n = lp->port_id;

	spin_lock(&lp->rx_lock);
	status = axienet_dma_in32( lp, n, XAXIDMA_RX_SR_OFFSET);

	while( (status & (XAXIDMA_IRQ_IOC_MASK | XAXIDMA_IRQ_DELAY_MASK)) && (work_done < quota)) {

#ifdef _DEBUG_
                printk("[%s]DBG> RX_DMASR:0x%08X, work_done:%d.\n", __func__, status, work_done );
                printk("[%s]DBG>(IOC_IRQ=0x1000, DLY_IRQ=0x2000) \n", __func__ );
#endif

		axienet_dma_out32( lp, n, XAXIDMA_RX_SR_OFFSET, status);
		if (status & XAXIDMA_IRQ_ERROR_MASK) {
			dev_err(&pdrv->pcidev->dev, "Rx error 0x%x ch(%d)\n\r", status, n);
			break;
		}
		work_done += axienet_recv( lp, quota - work_done);
		status = axienet_dma_in32( lp, n, XAXIDMA_RX_SR_OFFSET);
	}
	spin_unlock(&lp->rx_lock);

	/* Enable the interrupts again */
	cr = axienet_dma_in32(lp, n, XAXIDMA_RX_CR_OFFSET);
	cr |= (XAXIDMA_IRQ_IOC_MASK | XAXIDMA_IRQ_DELAY_MASK);
	axienet_dma_out32(lp, n, XAXIDMA_RX_CR_OFFSET, cr);

	wake_up_interruptible( &lp ->rx_pkt_q );

	return ;
}



void __axienet_device_reset( struct pcidrv_port *lp, off_t offset)
{
	struct pcidrv *pdrv = lp->pcidrv;
	int n = lp->port_id;

	u32 timeout;

/*	Reset Axi DMA. This would reset Axi Ethernet core as well. 
	The reset process of Axi DMA takes a while to complete.
	All pending commands/transfers will be flushed or completed during this reset process. */

	axienet_dma_out32( lp, n, offset, XAXIDMA_CR_RESET_MASK);
	timeout = 1000;

	while( axienet_dma_in32(lp, n, offset) & XAXIDMA_CR_RESET_MASK ) {
		udelay(1);
		if (--timeout == 0) {
			dev_err( &pdrv->pcidev->dev, "%s: DMA reset timeout!\n", __func__);
			break;
		}
	}
}



void axienet_dma_err_handler( unsigned long data)
{
	u32 cr, i;
	struct pcidrv_port *lp = (struct pcidrv_port *)data;
//	struct pcidrv *pdrv = lp->pcidrv;
	int n = lp->port_id;

	struct axidma_bd *cur_p;

	printk("__%s TODO PORT RESET\n",__FUNCTION__);
	//TODO PORT DISABLE

	//DMA RESET
	__axienet_device_reset(lp, XAXIDMA_TX_CR_OFFSET);
	__axienet_device_reset(lp, XAXIDMA_RX_CR_OFFSET);

	//DMA CLEAR
	for (i = 0; i < TX_BD_NUM; i++) {
		cur_p = &lp->tx_bd_v[i];
		cur_p->cntrl = 0;
		cur_p->status = 0;
		cur_p->app0 = 0;
		cur_p->app1 = 0;
		cur_p->app2 = 0;
		cur_p->app3 = 0;
		cur_p->app4 = 0;
	}

	for (i = 0; i < RX_BD_NUM; i++) {
		cur_p = &lp->rx_bd_v[i];
		cur_p->status = 0;
		cur_p->app0 = 0;
		cur_p->app1 = 0;
		cur_p->app2 = 0;
		cur_p->app3 = 0;
		cur_p->app4 = 0;
	}

	//TODO BUF FREE
	for( i=0; i < MAX_RX_BUFS; i++)
		if( lp->rx_data[i].data )	kfree( lp->rx_data[i].data );

	lp->tx_bd_ci = 0;
	lp->tx_bd_tail = 0;
	lp->rx_bd_ci = 0;

	/* Start updating the Rx channel control register */
	cr = axienet_dma_in32( lp, n, XAXIDMA_RX_CR_OFFSET);

	/* Update the interrupt coalesce count */
	cr = ( (cr & ~XAXIDMA_COALESCE_MASK) | (XAXIDMA_DFT_RX_THRESHOLD << XAXIDMA_COALESCE_SHIFT) );
	/* Update the delay timer count */
	cr = ( (cr & ~XAXIDMA_DELAY_MASK) | (XAXIDMA_DFT_RX_WAITBOUND << XAXIDMA_DELAY_SHIFT) );
	/* Enable coalesce, delay timer and error interrupts */
	cr |= XAXIDMA_IRQ_ALL_MASK;
	/* Finally write to the Rx channel control register */
	axienet_dma_out32( lp, n, XAXIDMA_RX_CR_OFFSET, cr);

	/* Start updating the Tx channel control register */
	cr = axienet_dma_in32( lp, n, XAXIDMA_TX_CR_OFFSET);

	/* Update the interrupt coalesce count */
	cr = ( (cr & ~XAXIDMA_COALESCE_MASK) | (XAXIDMA_DFT_TX_THRESHOLD << XAXIDMA_COALESCE_SHIFT) );
	/* Update the delay timer count */
	cr = ( ((cr & ~XAXIDMA_DELAY_MASK)) | (XAXIDMA_DFT_TX_WAITBOUND << XAXIDMA_DELAY_SHIFT) );
	/* Enable coalesce, delay timer and error interrupts */
	cr |= XAXIDMA_IRQ_ALL_MASK;
	/* Finally write to the Tx channel control register */
	axienet_dma_out32( lp, n, XAXIDMA_TX_CR_OFFSET, cr);

	/* Populate the tail pointer and bring the Rx Axi DMA engine out of halted state.
	 * This will make the Rx side ready for reception. */
	axienet_dma_bdout( lp, n, XAXIDMA_RX_CDESC_OFFSET, lp->rx_bd_p);
	cr = axienet_dma_in32( lp, n, XAXIDMA_RX_CR_OFFSET);

	axienet_dma_out32( lp, n, XAXIDMA_RX_CR_OFFSET, cr | XAXIDMA_CR_RUNSTOP_MASK );
	axienet_dma_bdout( lp, n, XAXIDMA_RX_TDESC_OFFSET, 
			lp->rx_bd_p + (sizeof(*lp->rx_bd_v) * (RX_BD_NUM-1)) );

	/* Write to the RS (Run-stop) bit in the Tx channel control register.
	 * Tx channel is now ready to run. But only after we write to the tail pointer register
	 * that the Tx channel will start transmitting */
	axienet_dma_bdout( lp, n, XAXIDMA_TX_CDESC_OFFSET, lp->tx_bd_p);
	cr = axienet_dma_in32( lp, n, XAXIDMA_TX_CR_OFFSET);
	axienet_dma_out32( lp, n, XAXIDMA_TX_CR_OFFSET, cr | XAXIDMA_CR_RUNSTOP_MASK );
}
#endif

