/*
 * Header file for AXI DMA.
 */

#ifndef _AIXDMA_H_
#define _AIXDMA_H_

/* axienet_dma_in32 - reads Memory mapped Axi DMA register.
 * This function returns the contents of the corresponding Axi DMA register. */

static inline u32 axienet_dma_in32(struct pcidrv_port *lp, int channel, off_t reg) {
	return __raw_readl(lp->bar + (channel + 1) * 0x10000 + reg);
}


/* axienet_dma_out32 - writes Memory mapped Axi DMA register.
 * This function writes the desired value into the corresponding Axi DMA register. */

static inline void axienet_dma_out32( struct pcidrv_port *lp, int channel, off_t reg, u32 value) {
	__raw_writel( value, (lp->bar + (channel + 1) * 0x10000 + reg));
}


/* axienet_dma_bdout - writes Memory mapped Axi DMA register Buffer Descriptor.
 * This function writes the desired value into the corresponding Axi DMA register. */

static inline void axienet_dma_bdout( struct pcidrv_port *lp, int channel, off_t reg, dma_addr_t value)
{
#if defined(CONFIG_PHYS_ADDR_T_64BIT)
	writeq( value, (lp->bar + (channel+1) * 0x10000 + reg) );
#else
	writel( value, (lp->bar + (channel+1) * 0x10000 + reg) );
#endif
}


/* This function reads the contents from the corresponding Axi DMA register. */

static inline u32 axi_pci_read( struct pcidrv *pcidrv, off_t reg) {
	 return __raw_readl( pcidrv->bar + reg );
}


static inline void axi_pci_write( struct pcidrv *pcidrv, off_t reg, u32 value) {
	 __raw_writel ( value, pcidrv->bar + reg);
}


void __axienet_device_reset( struct pcidrv_port *lp, off_t offset);
int  axienet_dma_bd_init( struct pcidrv_port *);
void axienet_dma_bd_release( struct pcidrv_port*);
int  axienet_start_xmit( struct pcidrv_port *lp, const char *buff, uint32_t len);
int  axienet_tx_irq(void *);
int  axienet_rx_irq(void *);
void axienet_dma_err_handler( unsigned long data);

void xaxienet_rx_poll( unsigned long data);
void xaxienet_tx_poll( unsigned long data);

#ifdef _DEBUG_
void print_bd( struct axidma_bd *bdp );
#endif

/* Axi DMA Register definitions.
 * Refer to (PG021) LogiCORE IP AXI DMA v7.1 */

#define XAXIDMA_TX_CR_OFFSET	0x00000000 /* Channel control */
#define XAXIDMA_TX_SR_OFFSET	0x00000004 /* Status */
#define XAXIDMA_TX_CDESC_OFFSET 0x00000008 /* Current descriptor pointer */
#define XAXIDMA_TX_TDESC_OFFSET 0x00000010 /* Tail descriptor pointer */

#define XAXIDMA_RX_CR_OFFSET	0x00000030 /* Channel control */
#define XAXIDMA_RX_SR_OFFSET	0x00000034 /* Status */
#define XAXIDMA_RX_CDESC_OFFSET 0x00000038 /* Current descriptor pointer */
#define XAXIDMA_RX_TDESC_OFFSET 0x00000040 /* Tail descriptor pointer */

#define XAXIDMA_CR_RUNSTOP_MASK 0x00000001 /* Start/stop DMA channel */
#define XAXIDMA_CR_RESET_MASK	0x00000004 /* Reset DMA engine */

#define XAXIDMA_BD_NDESC_OFFSET		0x00 /* Next descriptor pointer */
#define XAXIDMA_BD_BUFA_OFFSET		0x08 /* Buffer address */
#define XAXIDMA_BD_CTRL_LEN_OFFSET	0x18 /* Control/buffer length */
#define XAXIDMA_BD_STS_OFFSET		0x1C /* Status */
#define XAXIDMA_BD_USR0_OFFSET		0x20 /* User IP specific word0 */
#define XAXIDMA_BD_USR1_OFFSET		0x24 /* User IP specific word1 */
#define XAXIDMA_BD_USR2_OFFSET		0x28 /* User IP specific word2 */
#define XAXIDMA_BD_USR3_OFFSET		0x2C /* User IP specific word3 */
#define XAXIDMA_BD_USR4_OFFSET		0x30 /* User IP specific word4 */
#define XAXIDMA_BD_ID_OFFSET		0x34 /* Sw ID */
#define XAXIDMA_BD_HAS_STSCNTRL_OFFSET	0x38 /* Whether has stscntrl strm */
#define XAXIDMA_BD_HAS_DRE_OFFSET	0x3C /* Whether has DRE */

#define XAXIDMA_BD_HAS_DRE_SHIFT	8 /* Whether has DRE shift */
#define XAXIDMA_BD_HAS_DRE_MASK		0xF00 /* Whether has DRE mask */
#define XAXIDMA_BD_WORDLEN_MASK		0xFF /* Whether has DRE mask */

#define XAXIDMA_BD_CTRL_LENGTH_MASK	0x007FFFFF /* Requested len */
#define XAXIDMA_BD_CTRL_TXSOF_MASK	0x08000000 /* First tx packet */
#define XAXIDMA_BD_CTRL_TXEOF_MASK	0x04000000 /* Last tx packet */
#define XAXIDMA_BD_CTRL_ALL_MASK	0x0C000000 /* All control bits */

#define XAXIDMA_DELAY_MASK		0xFF000000 /* Delay timeout counter */
#define XAXIDMA_COALESCE_MASK		0x00FF0000 /* Coalesce counter */

#define XAXIDMA_DELAY_SHIFT		24
#define XAXIDMA_COALESCE_SHIFT		16

#define XAXIDMA_IRQ_IOC_MASK		0x00001000 /* Completion intr */
#define XAXIDMA_IRQ_DELAY_MASK		0x00002000 /* Delay interrupt */
#define XAXIDMA_IRQ_ERROR_MASK		0x00004000 /* Error interrupt */
#define XAXIDMA_IRQ_ALL_MASK		0x00007000 /* All interrupts */


/* Default TX/RX Threshold and waitbound values for SGDMA mode. */
#define XAXIDMA_DFT_TX_THRESHOLD	1
#define XAXIDMA_DFT_TX_WAITBOUND	1
#define XAXIDMA_DFT_RX_THRESHOLD	1
#define XAXIDMA_DFT_RX_WAITBOUND	1

/* old
#define XAXIDMA_DFT_TX_THRESHOLD	24
#define XAXIDMA_DFT_TX_WAITBOUND	254
#define XAXIDMA_DFT_RX_THRESHOLD	1
#define XAXIDMA_DFT_RX_WAITBOUND	128
*/

#define XAXIDMA_BD_STS_ACTUAL_LEN_MASK	0x007FFFFF /* Actual len */
#define XAXIDMA_BD_STS_COMPLETE_MASK	0x80000000 /* Completed */
#define XAXIDMA_BD_STS_DEC_ERR_MASK	0x40000000 /* Decode error */
#define XAXIDMA_BD_STS_SLV_ERR_MASK	0x20000000 /* Slave error */
#define XAXIDMA_BD_STS_INT_ERR_MASK	0x10000000 /* Internal err */
#define XAXIDMA_BD_STS_ALL_ERR_MASK	0x70000000 /* All errors */
#define XAXIDMA_BD_STS_RXSOF_MASK	0x08000000 /* First rx pkt */
#define XAXIDMA_BD_STS_RXEOF_MASK	0x04000000 /* Last rx pkt */
#define XAXIDMA_BD_STS_ALL_MASK		0xFC000000 /* All status bits */

#endif

